/*$Id: mgtype.c,v 1.20 2000/01/11 21:00:31 bsmith Exp bsmith $*/

#include "sys.h"
#include "mat.h"       /*I "mat.h"  I*/

#undef __FUNC__  
#define __FUNC__ "MatGetTypeFromOptions"
/*@C
   MatGetTypeFromOptions - Determines from the options database what matrix
   format the user has specified.

   Collective over MPI_Comm

   Input Parameter:
+  comm - the MPI communicator
.  type - the type of matrix desired, for example MATSEQAIJ, MATMPIAIJ
-  pre - optional string to prepend to the name

   Output Parameters:
.  set - flag indicating whether user set matrix type option; or PETSC_NULL.

   Basic Options Database Keys:
   These options return MATSEQxxx or MATMPIxxx,
   depending on the communicator, comm.
+    -mat_aij      - AIJ type
.    -mat_baij     - block AIJ type
.    -mat_dense    - dense type
-    -mat_bdiag    - block diagonal type

   More Options Database Keys:
+    -mat_seqaij   - MATSEQAIJ
.    -mat_mpiaij   - MATMPIAIJ
.    -mat_seqbaij  - MATSEQBAIJ
.    -mat_mpibaij  - MATMPIBAIJ
.    -mat_seqbdiag - MATSEQBDAIG
.    -mat_mpibdiag - MATMPIBDIAG
.    -mat_mpirowbs - MATMPIROWBS
.    -mat_seqdense - MATSEQDENSE
-    -mat_mpidense - MATMPIDENSE

   Notes:
   This routine is automatically called within MatCreate() and MatLoad().

   Level: developer

.keywords: matrix, get, format, from, options

.seealso: MatCreate(), MatLoad()
@*/

int MatGetTypeFromOptions(MPI_Comm comm,char *pre,MatType *type,PetscTruth *set)
{
  int               size,ierr;
  PetscTruth        flg1,flg2,flg3,flg4,flg5,flg8,flg9,flg10,flg12,flg13,flg11,flg14,flg15,flg16;
  static PetscTruth helpprinted = PETSC_FALSE;
  char              p[64];

  PetscFunctionBegin;
  PetscValidIntPointer(type); if (set) PetscValidIntPointer(set);

  ierr = PetscStrcpy(p,"-");CHKERRQ(ierr);
  if (pre) {ierr = PetscStrcat(p,pre);CHKERRQ(ierr);}

  ierr = MPI_Comm_size(comm,&size);CHKERRQ(ierr);
  ierr = OptionsHasName(PETSC_NULL,"-help",&flg1);CHKERRQ(ierr);
  if (flg1 && !helpprinted) {
    ierr = (*PetscHelpPrintf)(comm,"Matrix format options:\n");CHKERRQ(ierr);
    ierr = (*PetscHelpPrintf)(comm,"  %smat_aij, %smat_seqaij, %smat_mpiaij\n",p,p,p);CHKERRQ(ierr);
    ierr = (*PetscHelpPrintf)(comm,"  %smat_baij, %smat_seqbaij, %smat_mpibaij\n",p,p,p);CHKERRQ(ierr);
    ierr = (*PetscHelpPrintf)(comm,"  %smat_dense, %smat_seqdense, %smat_mpidense\n",p,p,p);CHKERRQ(ierr);
    ierr = (*PetscHelpPrintf)(comm,"  %smat_mpirowbs, %smat_bdiag, %smat_seqbdiag, %smat_mpibdiag\n",p,p,p,p);CHKERRQ(ierr);
    helpprinted = PETSC_TRUE;
  }
  ierr = OptionsHasName(pre,"-mat_seqdense",&flg1);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_mpidense",&flg2);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_seqbdiag",&flg3);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_mpibdiag",&flg4);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_mpirowbs",&flg5);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_seqbaij",&flg11);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_mpibaij",&flg14);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_mpiaij",&flg8);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_seqaij",&flg9);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_aij",&flg10);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_bdiag",&flg12);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_dense",&flg13);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_baij",&flg15);CHKERRQ(ierr);
  ierr = OptionsHasName(pre,"-mat_csr",&flg16);CHKERRQ(ierr);
  if (flg1) {
    *type = MATSEQDENSE;
    if (set) *set = PETSC_TRUE;
  }
  else if (flg2) {
    *type = MATMPIDENSE;
    if (set) *set = PETSC_TRUE;
  }
  else if (flg3) {
    *type = MATSEQBDIAG;
    if (set) *set = PETSC_TRUE;
  }
  else if (flg4) {
    *type = MATMPIBDIAG;
    if (set) *set = PETSC_TRUE;
  }
  else if (flg5) {
    *type = MATMPIROWBS;
    if (set) *set = PETSC_TRUE;
  }
  else if (flg8) {
    *type = MATMPIAIJ;
    if (set) *set = PETSC_TRUE;
  }
  else if (flg9){
    *type = MATSEQAIJ;
    if (set) *set = PETSC_TRUE;
  }
  else if (flg10) {
    if (size == 1) *type = MATSEQAIJ;
    else *type = MATMPIAIJ;
    if (set) *set = PETSC_TRUE;
  }  
  else if (flg11) {
    *type = MATSEQBAIJ;
    if (set) *set = PETSC_TRUE;
  }
  else if (flg12) {
    if (size == 1) *type = MATSEQBDIAG;
    else *type = MATMPIBDIAG;
    if (set) *set = PETSC_TRUE;
  }  
  else if (flg13) {
    if (size == 1) *type = MATSEQDENSE;
    else *type = MATMPIDENSE;
    if (set) *set = PETSC_TRUE;
  } 
  else if (flg14) {
    *type = MATMPIBAIJ;
    if (set) *set = PETSC_TRUE;
  }  
  else if (flg15) {
    if (size == 1) *type = MATSEQBAIJ;
    else *type = MATMPIBAIJ;
    if (set) *set = PETSC_TRUE;
  }  
  else if (flg16) {
    *type = MATMPIADJ;
    if (set) *set = PETSC_TRUE;
  }  
  else {
    if (size == 1) *type = MATSEQAIJ;
    else *type = MATMPIAIJ;
    if (set) *set = PETSC_FALSE;
  }
  PetscFunctionReturn(0);
}
